<?php

namespace Botble\Careers\Services;

use Botble\Base\Enums\BaseStatusEnum;
use Botble\Base\Supports\Helper;
use Botble\Careers\Models\Category;
use Botble\Careers\Models\Career;
use Botble\Careers\Models\Tag;
use Botble\Careers\Repositories\Interfaces\CategoryInterface;
use Botble\Careers\Repositories\Interfaces\CareerInterface;
use Botble\Careers\Repositories\Interfaces\TagInterface;
use Botble\SeoHelper\SeoOpenGraph;
use Botble\Slug\Models\Slug;
use Eloquent;
use Illuminate\Support\Arr;
use Illuminate\Support\Facades\Auth;
use RvMedia;
use SeoHelper;
use Theme;

class CareersService
{
    /**
     * @param Slug $slug
     * @return array|Eloquent
     */
    public function handleFrontRoutes($slug)
    {
        if (!$slug instanceof Eloquent) {
            return $slug;
        }

        $condition = [
            'id'     => $slug->reference_id,
            'status' => BaseStatusEnum::PUBLISHED,
        ];

        if (Auth::check() && request()->input('preview')) {
            Arr::forget($condition, 'status');
        }

        switch ($slug->reference_type) {
            case Career::class:
                $career = app(CareerInterface::class)
                    ->getFirstBy($condition, ['*'],
                        ['categories', 'tags', 'slugable', 'categories.slugable', 'tags.slugable']);

                if (empty($career)) {
                    abort(404);
                }

                Helper::handleViewCount($career, 'viewed_careersposts');

                SeoHelper::setTitle($career->name)
                    ->setDescription($career->description);

                $meta = new SeoOpenGraph;
                if ($career->image) {
                    $meta->setImage(RvMedia::getImageUrl($career->image));
                }
                $meta->setDescription($career->description);
                $meta->setUrl($career->url);
                $meta->setTitle($career->name);
                $meta->setType('article');

                SeoHelper::setSeoOpenGraph($meta);

                if (function_exists('admin_bar') && Auth::check() && Auth::user()->hasPermission('careersposts.edit')) {
                    admin_bar()->registerLink(trans('plugins/careers::careersposts.edit_this_careersposts'),
                        route('careersposts.edit', $career->id));
                }

                Theme::breadcrumb()->add(__('Home'), route('public.index'));

                $category = $career->categories->sortByDesc('id')->first();
                if ($category) {

                    if ($category->parents->count()) {
                        foreach ($category->parents as $parentCategory) {
                            Theme::breadcrumb()->add($parentCategory->name, $parentCategory->url);
                        }
                    }

                    Theme::breadcrumb()->add($category->name, $category->url);
                }

                Theme::breadcrumb()->add(SeoHelper::getTitle(), $career->url);

                do_action(BASE_ACTION_PUBLIC_RENDER_SINGLE, CAREER_MODULE_SCREEN_NAME, $career);

                return [
                    'view'         => 'career',
                    'default_view' => 'plugins/careers::themes.career',
                    'data'         => compact('career'),
                    'slug'         => $career->slug,
                ];
            case Category::class:
                $category = app(CategoryInterface::class)
                    ->getFirstBy($condition, ['*'], ['slugable']);

                if (empty($category)) {
                    abort(404);
                }

                SeoHelper::setTitle($category->name)
                    ->setDescription($category->description);

                $meta = new SeoOpenGraph;
                if ($category->image) {
                    $meta->setImage(RvMedia::getImageUrl($category->image));
                }
                $meta->setDescription($category->description);
                $meta->setUrl($category->url);
                $meta->setTitle($category->name);
                $meta->setType('article');

                SeoHelper::setSeoOpenGraph($meta);

                if (function_exists('admin_bar') && Auth::check() && Auth::user()->hasPermission('categories.edit')) {
                    admin_bar()->registerLink(trans('plugins/careers::categories.edit_this_category'),
                        route('categories.edit', $category->id));
                }

                $allRelatedCategoryIds = $category->getChildrenIds($category, [$category->id]);

                $careersposts = app(CareerInterface::class)
                    ->getByCategory($allRelatedCategoryIds, (int)theme_option('number_of_careers_in_a_category', 12));

                Theme::breadcrumb()
                    ->add(__('Home'), route('public.index'));

                if ($category->parents->count()) {
                    foreach ($category->parents->reverse() as $parentCategory) {
                        Theme::breadcrumb()->add($parentCategory->name, $parentCategory->url);
                    }
                }

                Theme::breadcrumb()->add(SeoHelper::getTitle(), $category->url);

                do_action(BASE_ACTION_PUBLIC_RENDER_SINGLE, CATEGORY_MODULE_SCREEN_NAME, $category);

                return [
                    'view'         => 'category',
                    'default_view' => 'plugins/careers::themes.category',
                    'data'         => compact('category', 'careersposts'),
                    'slug'         => $category->slug,
                ];
            case Tag::class:
                $tag = app(TagInterface::class)->getFirstBy($condition, ['*'], ['slugable']);

                if (!$tag) {
                    abort(404);
                }

                SeoHelper::setTitle($tag->name)
                    ->setDescription($tag->description);

                $meta = new SeoOpenGraph;
                $meta->setDescription($tag->description);
                $meta->setUrl($tag->url);
                $meta->setTitle($tag->name);
                $meta->setType('article');

                if (function_exists('admin_bar') && Auth::check() && Auth::user()->hasPermission('tags.edit')) {
                    admin_bar()->registerLink(trans('plugins/careers::tags.edit_this_tag'), route('tags.edit', $tag->id));
                }

                $careersposts = get_careersposts_by_tag($tag->id, (int)theme_option('number_of_careersposts_in_a_tag', 12));

                Theme::breadcrumb()
                    ->add(__('Home'), route('public.index'))
                    ->add(SeoHelper::getTitle(), $tag->url);

                do_action(BASE_ACTION_PUBLIC_RENDER_SINGLE, TAG_MODULE_SCREEN_NAME, $tag);

                return [
                    'view'         => 'tag',
                    'default_view' => 'plugins/careers::themes.tag',
                    'data'         => compact('tag', 'careersposts'),
                    'slug'         => $tag->slug,
                ];
        }

        return $slug;
    }
}
