<?php 

use Botble\Base\Enums\BaseStatusEnum;
use Botble\Base\Models\BaseModel;
use Botble\Base\Supports\SortItemsWithChildrenHelper;
use Botble\Careers\Repositories\Interfaces\CategoryInterface;
use Botble\Careers\Repositories\Interfaces\CareerInterface;
use Botble\Careers\Repositories\Interfaces\TagInterface;
use Botble\Careers\Supports\CareerFormat;
use Illuminate\Support\Arr;
use Illuminate\Support\Collection;

if (!function_exists('get_featured_careersposts')) {
    /**
     * @param int $limit
     * @param array $with
     * @return Collection
     */
    function get_featured_careersposts(int $limit, array $with = []): Collection
    {
        return app(CareerInterface::class)->getFeatured($limit, $with);
    }
}

if (!function_exists('get_latest_careersposts')) {
    /**
     * @param int $limit
     * @param array $excepts
     * @param array $with
     * @return Collection
     */
    function get_latest_careersposts(int $limit, array $excepts = [], array $with = []): Collection
    {
        return app(CareerInterface::class)->getListCareerNonInList($excepts, $limit, $with);
    }
}

if (!function_exists('get_related_careersposts')) {
    /**
     * @param int $id
     * @param int $limit
     * @return Collection
     */
    function get_related_careersposts(int $id, int $limit): Collection
    {
        return app(CareerInterface::class)->getRelated($id, $limit);
    }
}

if (!function_exists('get_careersposts_by_category')) {
    /**
     * @param int $categoryId
     * @param int $paginate
     * @param int $limit
     * @return Collection
     */
    function get_careersposts_by_category(int $categoryId, int $paginate = 12, int $limit = 0)
    {
        return app(CareerInterface::class)->getByCategory($categoryId, $paginate, $limit);
    }
}

if (!function_exists('get_careersposts_by_tag')) {
    /**
     * @param string $slug
     * @param int $paginate
     * @return Collection
     */
    function get_careersposts_by_tag(string $slug, int $paginate = 12)
    {
        return app(CareerInterface::class)->getByTag($slug, $paginate);
    }
}

if (!function_exists('get_careersposts_by_user')) {
    /**
     * @param int $authorId
     * @param int $paginate
     * @return Collection
     */
    function get_careersposts_by_user(int $authorId, int $paginate = 12)
    {
        return app(CareerInterface::class)->getByUserId($authorId, $paginate);
    }
}

if (!function_exists('get_all_careersposts')) {
    /**
     * @param boolean $active
     * @param int $perPage
     * @param array $with
     * @return Collection
     */
    function get_all_careersposts(
        bool  $active = true,
        int   $perPage = 12,
        array $with = ['slugable', 'categories', 'categories.slugable', 'author']
    ) {
        return app(CareerInterface::class)->getAllCareersposts($perPage, $active, $with);
    }
}

if (!function_exists('get_recent_careersposts')) {
    /**
     * @param int $limit
     * @return Collection
     */
    function get_recent_careersposts(int $limit, $catId = null)
    {
        return app(CareerInterface::class)->getRecentCareersposts($limit, $catId);
    }
}

if (!function_exists('get_featured_categories')) {
    /**
     * @param int $limit
     * @param array $with
     * @return Collection
     */
    function get_featured_categories(int $limit, array $with = [])
    {
        return app(CategoryInterface::class)->getFeaturedCategories($limit, $with);
    }
}

if (!function_exists('get_all_categories')) {
    /**
     * @param array $condition
     * @param array $with
     * @return Collection
     */
    function get_all_categories(array $condition = [], array $with = [])
    {
        return app(CategoryInterface::class)->getAllCategories($condition, $with);
    }
}

if (!function_exists('get_all_tags')) {
    /**
     * @param boolean $active
     * @return Collection
     */
    function get_all_tags(bool $active = true)
    {
        return app(TagInterface::class)->getAllTags($active);
    }
}

if (!function_exists('get_popular_tags')) {
    /**
     * @param int $limit
     * @param array|string[] $with
     * @param array $withCount
     * @return Collection
     */
    function get_popular_tags(int $limit = 10, array $with = ['slugable'], array $withCount = ['careersposts'])
    {
        return app(TagInterface::class)->getPopularTags($limit, $with, $withCount);
    }
}

if (!function_exists('get_popular_careersposts')) {
    /**
     * @param integer $limit
     * @param array $args
     * @return Collection
     */
    function get_popular_careersposts(int $limit = 10, array $args = [])
    {
        return app(CareerInterface::class)->getPopularCareersposts($limit, $args);
    }
}

if (!function_exists('get_popular_categories')) {
    /**
     * @param integer $limit
     * @param array $with
     * @param array $withCount
     * @return Collection
     */
    function get_popular_categories(int $limit = 10, array $with = ['slugable'], array $withCount = ['careersposts'])
    {
        return app(CategoryInterface::class)->getPopularCategories($limit, $with, $withCount);
    }
}

if (!function_exists('get_category_by_id')) {
    /**
     * @param integer $id
     * @return BaseModel
     */
    function get_category_by_id(int $id): BaseModel
    {
        return app(CategoryInterface::class)->getCategoryById($id);
    }
}

if (!function_exists('get_categories')) {
    /**
     * @param array $args
     * @return array
     */
    function get_categories(array $args = []): array
    {
        $indent = Arr::get($args, 'indent', '——');

        $repo = app(CategoryInterface::class);

        $categories = $repo->getCategories(Arr::get($args, 'select', ['*']), [
            'created_at' => 'DESC',
            'is_default' => 'DESC',
            'order'      => 'ASC',
        ]);

        $categories = sort_item_with_children($categories);

        foreach ($categories as $category) {
            $depth = (int)$category->depth;
            $indentText = str_repeat($indent, $depth);
            $category->indent_text = $indentText;
        }

        return $categories;
    }
}

if (!function_exists('get_categories_with_children')) {
    /**
     * @return Collection
     * @throws Exception
     */
    function get_categories_with_children()
    {
        $categories = app(CategoryInterface::class)
            ->getAllCategoriesWithChildren(['status' => BaseStatusEnum::PUBLISHED], [], ['id', 'name', 'parent_id']);

        return app(SortItemsWithChildrenHelper::class)
            ->setChildrenProperty('child_cats')
            ->setItems($categories)
            ->sort();
    }
}


if (!function_exists('get_newcategories_with_children')) {
    /**
     * @return Collection
     * @throws Exception
     */
    function get_newcategories_with_children()
    {
        $categories = app(CategoryInterface::class)
            ->getAllCategoriesWithChildren(['status' => BaseStatusEnum::PUBLISHED], [], ['id', 'name', 'parent_id']);

        return app(SortItemsWithChildrenHelper::class)
            ->setChildrenProperty('child_cats')
            ->setItems($categories)
            ->sort();
    }
}

if (!function_exists('register_careersposts_format')) {
    /**
     * @param array $formats
     * @return void
     */
    function register_careersposts_format(array $formats)
    {
        CareersFormat::registerCareersFormat($formats);
    }
}

if (!function_exists('get_careersposts_formats')) {
    /**
     * @param bool $convertToList
     * @return array
     */
    function get_careersposts_formats(bool $convertToList = false): array
    {
        return CareerFormat::getCareerFormats($convertToList);
    }
}

if (!function_exists('get_careersposts_page_id')) {
    /**
     * @return int
     */
    function get_careersposts_page_id(): int
    {
        return (int)theme_option('careersposts_page_id', setting('careersposts_page_id'));
    }
}

if (!function_exists('get_careersposts_page_url')) {
    /**
     * @return string
     */
    function get_careersposts_page_url(): string
    {
        $careersPageId = (int)theme_option('careersposts_page_id', setting('careersposts_page_id'));

        if (!$careersPageId) {
            return url('/');
        }

        $careersPage = app(\Botble\Page\Repositories\Interfaces\PageInterface::class)->findById($careersPageId);

        if (!$careersPage) {
            return url('/');
        }

        return $careersPage->url;
    }
}

if (!function_exists('get_recent_careersposts_by_category')) {
    /**
     * @param int $categoryId
     * @param int $paginate
     * @param int $limit
     * @return Collection
     */
    function get_recent_careersposts_by_category(array $categoryId,int $perPage=0,int $limit = 0)
    {
        return app(CareerInterface::class)->getRecentByCategory($categoryId,$perPage,$limit);
    }
}
if (!function_exists('get_child_category')) {
    /**
     * @param array $categoryId     
     * @param int $limit
     * @param int $paginate
     * @return Collection
     */
    function get_child_category(array $categoryId,int $perPage=0,int $limit = 0)
    {
        
        return app(CategoryInterface::class)->getChildCategory($categoryId,$perPage=0,$limit);
    }
}
if (!function_exists('get_category_details')) {
    /**
     * @param int $categoryId
     * @param int $paginate
     * @param int $limit
     * @return Collection
     */
    function get_category_details($categoryId)
    {
        return app(CategoryInterface::class)->getCategoryDetails($categoryId);
    }
}

if (!function_exists('get_next_careersposts')) {
    /**
     * @param int $categoryId
     * @param int $paginate
     * @param int $limit
     * @return Collection
     */
    function get_next_careersposts($categoryId,$postId)
    {
        return app(CareerInterface::class)->getNextCareer($categoryId,$careersId);
    }
}


if (!function_exists('get_prev_careersposts')) {
    /**
     * @param int $categoryId
     * @param int $paginate
     * @param int $limit
     * @return Collection
     */
    function get_prev_careersposts($categoryId,$careerstId)
    {
        return app(CareerInterface::class)->getPrevCareer($categoryId,$careersId);
    }
}
if (!function_exists('get_careersposts')) {
    /**
     * @param int $categoryId
     * @param int $paginate
     * @param int $limit
     * @return Collection
     */
    function get_careersposts($careersId)
    {
        return app(CareerInterface::class)->getCareer($careersId);
    }
}

if (!function_exists('get_career_posts')) {
    /**
     * Get posts in the "news" category.
     *
     * @param int $limit
     * @param array $with
     * @return Collection
     */
    function get_career_posts(int $limit = 12, array $with = []): Collection
    {
        return app(CareerInterface::class)
            ->getModel() // Assuming this method returns the Post model
            ->join('careers_categories', 'careers.id', '=', 'careers_categories.career_id')
            ->join('newcategories', 'careers_categories.category_id', '=', 'newcategories.id')
            ->select('careers.*')
            ->limit($limit)
            ->get();
    }
}

if (!function_exists('get_career_posts')) {
    /**
     * Get posts in the "news" category.
     *
     * @param int $limit
     * @param array $with
     * @return Collection
     */
    function get_career_single_posts(int $limit = 12, array $with = []): Collection
    {
        return app(CareerInterface::class)
            ->getModel() // Assuming this method returns the Post model
            ->join('careers_categories', 'careers.id', '=', 'careers_categories.career_id')
            ->join('newcategories', 'careers_categories.category_id', '=', 'newcategories.id')
            ->select('careers.*')
            ->where('careers.id',$id)
            ->limit($limit)
            ->get();
    }
}