<?php

namespace Botble\Careers\Http\Controllers;

use Botble\ACL\Models\User;
use Botble\Base\Events\BeforeEditContentEvent;
use Botble\Base\Events\CreatedContentEvent;
use Botble\Base\Events\DeletedContentEvent;
use Botble\Base\Events\UpdatedContentEvent;
use Botble\Base\Forms\FormBuilder;
use Botble\Base\Http\Controllers\BaseController;
use Botble\Base\Http\Responses\BaseHttpResponse;
use Botble\Base\Traits\HasDeleteManyItemsTrait;
use Botble\Careers\Forms\CareerForm;
use Botble\Careers\Http\Requests\CareerRequest;
use Botble\Careers\Models\Career;
use Botble\Careers\Repositories\Interfaces\CategoryInterface;
use Botble\Careers\Repositories\Interfaces\CareerInterface;
use Botble\Careers\Repositories\Interfaces\TagInterface;
use Botble\Careers\Services\StoreCategoryService;
use Botble\Careers\Services\StoreTagService;
use Botble\Careers\Tables\CareerTable;
use Exception;
use Illuminate\Contracts\View\Factory;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Auth;
use Illuminate\View\View;
use Throwable;

class CareerController extends BaseController
{

    use HasDeleteManyItemsTrait;

    /**
     * @var PostInterface
     */
    protected $careerRepository;

    /**
     * @var TagInterface
     */
    protected $tagRepository;

    /**
     * @var CategoryInterface
     */
    protected $categoryRepository;

    /**
     * @param PostInterface $postRepository
     * @param TagInterface $tagRepository
     * @param CategoryInterface $categoryRepository
     */
    public function __construct(
        CareerInterface $careerRepository,
        TagInterface $tagRepository,
        CategoryInterface $categoryRepository
    ) {
        $this->careerRepository = $careerRepository;
        $this->tagRepository = $tagRepository;
        $this->categoryRepository = $categoryRepository;
    }

    /**
     * @param PostTable $dataTable
     * @return Factory|View
     * @throws Throwable
     */
    public function index(CareerTable $dataTable)
    {
       
        page_title()->setTitle(trans('plugins/careers::documentsposts.menu_name'));

        return $dataTable->renderTable();
    }

    /**
     * @param FormBuilder $formBuilder
     * @return string
     */
    public function create(FormBuilder $formBuilder)
    {
        page_title()->setTitle(trans('Create'));

        return $formBuilder->create(CareerForm::class)->renderForm();
    }

    /**
     * @param PostRequest $request
     * @param StoreTagService $tagService
     * @param StoreCategoryService $categoryService
     * @param BaseHttpResponse $response
     * @return BaseHttpResponse
     */
    public function store(
        CareerRequest $request,
        StoreTagService $tagService,
        StoreCategoryService $categoryService,
        BaseHttpResponse $response
    ) {
        /**
         * @var Post $post
         */

      //  dd($request);
        $career = $this->careerRepository->createOrUpdate(array_merge($request->input(), [
            'author_id'   => Auth::id(),
            'author_type' => User::class,
        ]));
        

        
    //      $files = $request->file('uploaddocument');
    //     foreach ($files as $file) {
    //         $fileName = $file->getClientOriginalName(); // Get the original file name
    //         $file->move(public_path('uploaddocument'), $fileName); // Move each file to the desired location
    //     }

    //   print_r( $files );
    //  exit;

        event(new CreatedContentEvent(CAREER_MODULE_SCREEN_NAME, $request, $career));

        $tagService->execute($request, $career);

        $categoryService->execute($request, $career);

        return $response
            ->setPreviousUrl(route('careersposts.index'))
            ->setNextUrl(route('careersposts.edit', $career->id))
            ->setMessage(trans('core/base::notices.create_success_message'));

    }

    /**
     * @param int $id
     * @param FormBuilder $formBuilder
     * @param Request $request
     * @return string
     */
    public function edit($id, FormBuilder $formBuilder, Request $request)
    {
        $career = $this->careerRepository->findOrFail($id);

        event(new BeforeEditContentEvent($request, $career));

        page_title()->setTitle(trans('Edit') . ' "' . $career->name . '"');

        return $formBuilder->create(CareerForm::class, ['model' => $career])->renderForm();
    }

    /**
     * @param int $id
     * @param PostRequest $request
     * @param StoreTagService $tagService
     * @param StoreCategoryService $categoryService
     * @param BaseHttpResponse $response
     * @return BaseHttpResponse
     */
    public function update(
        $id,
        CareerRequest $request,
        StoreTagService $tagService,
        StoreCategoryService $categoryService,
        BaseHttpResponse $response
    ) {

       // dd($request);
        $career = $this->careerRepository->findOrFail($id);

        $career->fill($request->input());

        $this->careerRepository->createOrUpdate($career);

        event(new UpdatedContentEvent(CAREER_MODULE_SCREEN_NAME, $request, $career));

        $tagService->execute($request, $career);

        $categoryService->execute($request, $career);

        return $response
            ->setPreviousUrl(route('careersposts.index'))
            ->setMessage(trans('core/base::notices.update_success_message'));
    }

    /**
     * @param int $id
     * @param Request $request
     * @return BaseHttpResponse
     */
    public function destroy($id, Request $request, BaseHttpResponse $response)
    {
        try {
            $career = $this->careerRepository->findOrFail($id);
            $this->careerRepository->delete($career);

            event(new DeletedContentEvent(CAREER_MODULE_SCREEN_NAME, $request, $career));

            return $response
                ->setMessage(trans('core/base::notices.delete_success_message'));
        } catch (Exception $exception) {
            return $response
                ->setError()
                ->setMessage($exception->getMessage());
        }
    }

    /**
     * @param Request $request
     * @param BaseHttpResponse $response
     * @return BaseHttpResponse
     * @throws Exception
     */
    public function deletes(Request $request, BaseHttpResponse $response)
    {
        return $this->executeDeleteItems($request, $response, $this->careerRepository, CAREER_MODULE_SCREEN_NAME);
    }

    /**
     * @param Request $request
     * @param BaseHttpResponse $response
     * @return BaseHttpResponse
     * @throws Throwable
     */
    public function getWidgetRecentCareers(Request $request, BaseHttpResponse $response)
    {
        $limit = (int)$request->input('paginate', 10);
        $limit = $limit > 0 ? $limit : 10;

        $careers = $this->careersRepository->advancedGet([
            'with'     => ['slugable'],
            'order_by' => ['created_at' => 'desc'],
            'paginate' => [
                'per_page'      => $limit,
                'current_paged' => (int)$request->input('page', 1),
            ],
        ]);

        return $response
            ->setData(view('plugins/careers::careersposts.widgets.careersposts', compact('careersposts', 'limit'))->render());
    }
}
