<?php

namespace Botble\Documents\Services;

use Botble\Base\Enums\BaseStatusEnum;
use Botble\Base\Supports\Helper;
use Botble\Documents\Models\Category;
//use Botble\Documents\Models\Post;
use Botble\Documents\Models\Document;
use Botble\Documents\Models\Tag;
use Botble\Documents\Repositories\Interfaces\CategoryInterface;
//use Botble\Documents\Repositories\Interfaces\PostInterface;
use Botble\Documents\Repositories\Interfaces\DocumentInterface;
use Botble\Documents\Repositories\Interfaces\TagInterface;
use Botble\SeoHelper\SeoOpenGraph;
use Botble\Slug\Models\Slug;
use Eloquent;
use Illuminate\Support\Arr;
use Illuminate\Support\Facades\Auth;
use RvMedia;
use SeoHelper;
use Theme;

class DocumentsService
{
    /**
     * @param Slug $slug
     * @return array|Eloquent
     */
    public function handleFrontRoutes($slug)
    {
        if (!$slug instanceof Eloquent) {
            return $slug;
        }

        $condition = [
            'id'     => $slug->reference_id,
            'status' => BaseStatusEnum::PUBLISHED,
        ];

        if (Auth::check() && request()->input('preview')) {
            Arr::forget($condition, 'status');
        }

        switch ($slug->reference_type) {
            case Document::class:
                $document = app(DocumentInterface::class)
                    ->getFirstBy($condition, ['*'],
                        ['categories', 'tags', 'slugable', 'categories.slugable', 'tags.slugable']);

                if (empty($document)) {
                    abort(404);
                }

                Helper::handleViewCount($document, 'viewed_documentsposts');

                SeoHelper::setTitle($document->name)
                    ->setDescription($document->description);

                $meta = new SeoOpenGraph;
                if ($document->image) {
                    $meta->setImage(RvMedia::getImageUrl($document->image));
                }
                $meta->setDescription($document->description);
                $meta->setUrl($document->url);
                $meta->setTitle($document->name);
                $meta->setType('article');

                SeoHelper::setSeoOpenGraph($meta);

                if (function_exists('admin_bar') && Auth::check() && Auth::user()->hasPermission('documentsposts.edit')) {
                    admin_bar()->registerLink(trans('plugins/documents::documentsposts.edit_this_documentsposts'),
                        route('documentsposts.edit', $document->id));
                }

                Theme::breadcrumb()->add(__('Home'), route('public.index'));

                $category = $document->categories->sortByDesc('id')->first();
                if ($category) {

                    if ($category->parents->count()) {
                        foreach ($category->parents as $parentCategory) {
                            Theme::breadcrumb()->add($parentCategory->name, $parentCategory->url);
                        }
                    }

                    Theme::breadcrumb()->add($category->name, $category->url);
                }

                Theme::breadcrumb()->add(SeoHelper::getTitle(), $document->url);

                do_action(BASE_ACTION_PUBLIC_RENDER_SINGLE, DOCUMENT_MODULE_SCREEN_NAME, $document);

                return [
                    'view'         => 'document',
                    'default_view' => 'plugins/documents::themes.document',
                    'data'         => compact('document'),
                    'slug'         => $document->slug,
                ];
            case Category::class:
                $category = app(CategoryInterface::class)
                    ->getFirstBy($condition, ['*'], ['slugable']);

                if (empty($category)) {
                    abort(404);
                }

                SeoHelper::setTitle($category->name)
                    ->setDescription($category->description);

                $meta = new SeoOpenGraph;
                if ($category->image) {
                    $meta->setImage(RvMedia::getImageUrl($category->image));
                }
                $meta->setDescription($category->description);
                $meta->setUrl($category->url);
                $meta->setTitle($category->name);
                $meta->setType('article');

                SeoHelper::setSeoOpenGraph($meta);

                if (function_exists('admin_bar') && Auth::check() && Auth::user()->hasPermission('categories.edit')) {
                    admin_bar()->registerLink(trans('plugins/documents::categories.edit_this_category'),
                        route('categories.edit', $category->id));
                }

                $allRelatedCategoryIds = $category->getChildrenIds($category, [$category->id]);

                $documentsposts = app(DocumentInterface::class)
                    ->getByCategory($allRelatedCategoryIds, (int)theme_option('number_of_documents_in_a_category', 12));

                Theme::breadcrumb()
                    ->add(__('Home'), route('public.index'));

                if ($category->parents->count()) {
                    foreach ($category->parents->reverse() as $parentCategory) {
                        Theme::breadcrumb()->add($parentCategory->name, $parentCategory->url);
                    }
                }

                Theme::breadcrumb()->add(SeoHelper::getTitle(), $category->url);

                do_action(BASE_ACTION_PUBLIC_RENDER_SINGLE, CATEGORY_MODULE_SCREEN_NAME, $category);

                return [
                    'view'         => 'category',
                    'default_view' => 'plugins/documents::themes.category',
                    'data'         => compact('category', 'documentsposts'),
                    'slug'         => $category->slug,
                ];
            case Tag::class:
                $tag = app(TagInterface::class)->getFirstBy($condition, ['*'], ['slugable']);

                if (!$tag) {
                    abort(404);
                }

                SeoHelper::setTitle($tag->name)
                    ->setDescription($tag->description);

                $meta = new SeoOpenGraph;
                $meta->setDescription($tag->description);
                $meta->setUrl($tag->url);
                $meta->setTitle($tag->name);
                $meta->setType('article');

                if (function_exists('admin_bar') && Auth::check() && Auth::user()->hasPermission('tags.edit')) {
                    admin_bar()->registerLink(trans('plugins/documents::tags.edit_this_tag'), route('tags.edit', $tag->id));
                }

                $documentsposts = get_documentsposts_by_tag($tag->id, (int)theme_option('number_of_documentsposts_in_a_tag', 12));

                Theme::breadcrumb()
                    ->add(__('Home'), route('public.index'))
                    ->add(SeoHelper::getTitle(), $tag->url);

                do_action(BASE_ACTION_PUBLIC_RENDER_SINGLE, TAG_MODULE_SCREEN_NAME, $tag);

                return [
                    'view'         => 'tag',
                    'default_view' => 'plugins/documents::themes.tag',
                    'data'         => compact('tag', 'documentsposts'),
                    'slug'         => $tag->slug,
                ];
        }

        return $slug;
    }
}
