<?php

namespace Botble\Careers\Tables;

use BaseHelper;
use Botble\Base\Enums\BaseStatusEnum;
use Botble\Careers\Exports\CareerExport;
use Botble\Careers\Repositories\Interfaces\CategoryInterface;
use Botble\Careers\Repositories\Interfaces\CareerInterface;
use Botble\Table\Abstracts\TableAbstract;
use Html;
use Illuminate\Contracts\Routing\UrlGenerator;
use Illuminate\Support\Facades\Auth;
use Yajra\DataTables\DataTables;

class CareerTable extends TableAbstract
{
    /**
     * @var bool
     */
    protected $hasActions = true;

    /**
     * @var bool
     */
    protected $hasFilter = true;

    /**
     * @var CategoryInterface
     */
    protected $categoryRepository;

    /**
     * @var string
     */
    protected $exportClass = CareerExport::class;

    /**
     * @var int
     */
    protected $defaultSortColumn = 6;

    /**
     * PostTable constructor.
     * @param DataTables $table
     * @param UrlGenerator $urlGenerator
     * @param PostInterface $postRepository
     * @param CategoryInterface $categoryRepository
     */
    public function __construct(
        DataTables $table,
        UrlGenerator $urlGenerator,
        CareerInterface $careerRepository,
        CategoryInterface $categoryRepository
    ) {
        parent::__construct($table, $urlGenerator);

        $this->repository = $careerRepository;
        $this->categoryRepository = $categoryRepository;

        if (!Auth::user()->hasAnyPermission(['careersposts.edit', 'careersposts.destroy'])) {
            $this->hasOperations = false;
            $this->hasActions = false;
        }
    }

    /**
     * {@inheritDoc}
     */
    public function ajax()
    {
        $data = $this->table
            ->eloquent($this->query())
            ->editColumn('name', function ($item) {
                if (!Auth::user()->hasPermission('careersposts.edit')) {
                    return BaseHelper::clean($item->name);
                }

                return Html::link(route('careersposts.edit', $item->id), BaseHelper::clean($item->name));
            })
            // ->editColumn('image', function ($item) {
            //     return $this->displayThumbnail($item->image);
            // })
 
            ->editColumn('checkbox', function ($item) {
                return $this->getCheckbox($item->id);
            })
            ->editColumn('created_at', function ($item) {
                return BaseHelper::formatDate($item->created_at);
            })
            ->editColumn('updated_at', function ($item) {
                $categories = '';
                foreach ($item->categories as $category) {
                    $categories .= Html::link(route('categories.edit', $category->id), $category->name) . ', ';
                }

                return rtrim($categories, ', ');
            })
            // ->editColumn('author_id', function ($item) {
            //     return $item->author && $item->author->name ? BaseHelper::clean($item->author->name) : '&mdash;';
            // })

            ->editColumn('status', function ($item) {
                if ($this->request()->input('action') === 'excel') {
                    return $item->status->getValue();
                }

                return BaseHelper::clean($item->status->toHtml());
            })
          
              ->addColumn('operations', function ($item) {
                return $this->getOperations('careersposts.edit', 'careersposts.destroy', $item);
            })
            ;

            // ->addColumn('Actions', function ($item) {
            //     return $this->getOperations('careersposts.edit', 'careersposts.destroy', $item);
            // });

        return $this->toJson($data);
    }

    /**
     * {@inheritDoc}
     */
    public function query()
    {
        $query = $this->repository->getModel()
            ->with([
                'categories' => function ($query) {
                    $query->select(['newcategories.id', 'newcategories.name']);
                },
                'author',
            ])
            ->select([
                'id',
                'name',
                // 'image',
                'uploaddocument',
                'created_at',
                'status',
                'updated_at',
                // 'author_id',
                'author_type',
            ]);

        return $this->applyScopes($query);
    }

    /**
     * {@inheritDoc}
     */
    public function columns()
    {
        return [
            'id'         => [
                'title' => trans('core/base::tables.id'),
                'width' => '20px',
            ],
            // 'image'      => [
            //     'title' => trans('core/base::tables.image'),
            //     'width' => '70px',
            // ],
            
            'name'       => [
                'title' => trans('Job Title'),
                'class' => 'text-start',
            ],
            'updated_at' => [
                'title'     => trans('Categories'),
                'width'     => '150px',
                'class'     => 'no-sort text-center',
                'orderable' => false,
            ],
            // 'author_id'  => [
            //     'title'     => trans('Author'),
            //     'width'     => '150px',
            //     'class'     => 'no-sort text-center',
            //     'orderable' => false,
            // ],
            'created_at' => [
                'title' => trans('Created On'),
                'width' => '100px',
                'class' => 'text-center',
            ],
            'status'     => [
                'title' => trans('core/base::tables.status'),
                'width' => '100px',
                'class' => 'text-center',
            ],
        ];
    }

    /**
     * {@inheritDoc}
     */
    public function buttons()
    {
        return $this->addCreateButton(route('careersposts.create'), 'careersposts.create');
    }

    /**
     * {@inheritDoc}
     */
    public function bulkActions(): array
    {
        return $this->addDeleteAction(route('careersposts.deletes'), 'careersposts.destroy', parent::bulkActions());
    }

    /**
     * {@inheritDoc}
     */
    public function getBulkChanges(): array
    {
        return [
            'name'       => [
                'title'    => trans('Job Title'),
                'type'     => 'text',
                'validate' => 'required|max:120',
            ],
            
            'category'         => [
                'title'    => trans('Categories'),
                'type'     => 'select-search',
                'validate' => 'required',
                'callback' => 'getCategories',
            ],
            'created_at' => [
                'title'    => trans('core/base::tables.created_at'),
                'type'     => 'date',
                'validate' => 'required',
            ],
            'status'     => [
                'title'    => trans('core/base::tables.status'),
                'type'     => 'customSelect',
                'choices'  => BaseStatusEnum::labels(),
                'validate' => 'required|in:' . implode(',', BaseStatusEnum::values()),
            ],
           
        ];
    }

    /**
     * @return array
     */
    public function getCategories(): array
    {
        return $this->categoryRepository->pluck('name', 'id');
    }

    /**
     * {@inheritDoc}
     */
    public function applyFilterCondition($query, string $key, string $operator, ?string $value)
    {
        switch ($key) {
            case 'created_at':
                if (!$value) {
                    break;
                }

                $value = BaseHelper::formatDate($value);

                return $query->whereDate($key, $operator, $value);
            case 'category':
                if (!$value) {
                    break;
                }

                if (!BaseHelper::isJoined($query, 'careers_categories')) {
                    $query = $query
                        ->join('careers_categories', 'careers_categories.careers_id', '=', 'careers.id')
                        ->join('categories', 'careers_categories.category_id', '=', 'categories.id')
                        ->select($query->getModel()->getTable() . '.*');
                }

                return $query->where('careers_categories.category_id', $value);
        }

        return parent::applyFilterCondition($query, $key, $operator, $value);
    }

    /**
     * {@inheritDoc}
     */
    public function saveBulkChangeItem($item, string $inputKey, ?string $inputValue)
    {
        if ($inputKey === 'category') {
            $item->categories()->sync([$inputValue]);

            return $item;
        }

        return parent::saveBulkChangeItem($item, $inputKey, $inputValue);
    }

    /**
     * {@inheritDoc}
     */
    public function getDefaultButtons(): array
    {
        return [
            'export',
            'reload',
        ];
    }
}
